@@ -11,11 +11,27 @@ This source code is licensed under the MIT license found in the LICENSE
 file in the root directory of this source tree.
 """
 
+from typing import Any, Optional
+
 import numpy as np
 
+from sentinelhub.exceptions import deprecated_function
+
+from eolearn.core.exceptions import EODeprecationWarning
+
 
 # This code was copied from https://gist.github.com/seberg/3866040
-def rolling_window(array, window=(0,), asteps=None, wsteps=None, axes=None, toend=True):
+@deprecated_function(
+    category=EODeprecationWarning, message_suffix="Please use `numpy.lib.stride_tricks.sliding_window_view` instead."
+)
+def rolling_window(
+    array: np.ndarray,
+    window: Any = (0,),
+    asteps: Optional[Any] = None,
+    wsteps: Optional[Any] = None,
+    axes: Optional[Any] = None,
+    toend: bool = True,
+) -> np.ndarray:
     """Create a view of `array` which for every point gives the n-dimensional neighbourhood of size window. New
     dimensions are added at the end of `array` or after the corresponding original dimension.
 
@@ -72,25 +88,18 @@ def rolling_window(array, window=(0,), asteps=None, wsteps=None, axes=None, toen
                [5, 7, 9]])
 
     :param array: Array to which the rolling window is applied
-    :type array: np.ndarray
     :param window: Either a single integer to create a window of only the last axis or a tuple to create it for
         the last len(window) axes. 0 can be used to ignore a dimension in the window.
-    :type window: int or (int, int)
     :param asteps: Aligned at the last axis, new steps for the original array, i.e. for creation of non-overlapping
         windows. (Equivalent to slicing result)
-    :type asteps: tuple
     :param wsteps: Steps for the added window dimensions. These can be 0 to repeat values along the axis.
-    :type wsteps: int or tuple (same size as window) or None
     :param axes: If given, must have the same size as window. In this case window is interpreted as the size in the
         dimension given by axes. E.g. a window of (2, 1) is equivalent to window=2 and axis=-2.
-    :type axes: int or tuple
     :param toend: If False, the new dimensions are right after the corresponding original dimension, instead of at
         the end of the array. Adding the new axes at the end makes it easier to get the neighborhood, however
         toend=False will give a more intuitive result if you view the whole array.
-    :type toend: bool
     :returns: A view on `array` which is smaller to fit the windows and has windows added dimensions (0s not counting),
         i.e. every point of `array` is an array of size window.
-    :rtype: np.ndarray
     """
     # pylint: disable=too-many-branches
     # pylint: disable=too-many-statements
